use egui::{
    Align, CentralPanel, Color32, ComboBox, Context, Frame, Image, RichText, ScrollArea, Sense,
    Slider, TextEdit, TextureOptions, Ui, UiBuilder, Vec2, ViewportBuilder, ViewportId,
};
use egui_extras::{Column, Size, StripBuilder, TableBuilder};

use crate::{
    emulator::SimId,
    vram::{VramPalette, VramResource},
};

use super::AppWindow;

pub struct CharacterDataWindow {
    sim_id: SimId,
    palette: VramPalette,
    index: usize,
    index_str: String,
    scale: f32,
    show_grid: bool,
}

impl CharacterDataWindow {
    pub fn new(sim_id: SimId) -> Self {
        Self {
            sim_id,
            palette: VramPalette::Generic,
            index: 0,
            index_str: "0".into(),
            scale: 4.0,
            show_grid: true,
        }
    }

    fn show_form(&mut self, ui: &mut Ui) {
        let row_height = ui.spacing().interact_size.y;
        ui.vertical(|ui| {
            TableBuilder::new(ui)
                .column(Column::auto())
                .column(Column::remainder())
                .body(|mut body| {
                    body.row(row_height, |mut row| {
                        row.col(|ui| {
                            ui.label("Index");
                        });
                        row.col(|ui| {
                            let res = ui.add(
                                TextEdit::singleline(&mut self.index_str)
                                    .horizontal_align(Align::Max),
                            );
                            if res.changed() {
                                if let Some(index) =
                                    self.index_str.parse().ok().filter(|id| *id < 2048)
                                {
                                    self.index = index;
                                }
                            }
                        });
                    });
                    body.row(row_height, |mut row| {
                        row.col(|ui| {
                            ui.label("Address");
                        });
                        row.col(|ui| {
                            let address = match self.index {
                                0x000..0x200 => 0x00060000 + self.index,
                                0x200..0x400 => 0x000e0000 + (self.index - 0x200),
                                0x400..0x600 => 0x00160000 + (self.index - 0x400),
                                0x600..0x800 => 0x001e0000 + (self.index - 0x600),
                                _ => unreachable!("can't happen"),
                            };
                            let mut address_str = format!("{address:08x}");
                            ui.add_enabled(
                                false,
                                TextEdit::singleline(&mut address_str).horizontal_align(Align::Max),
                            );
                        });
                    });
                    body.row(row_height, |mut row| {
                        row.col(|ui| {
                            ui.label("Mirror");
                        });
                        row.col(|ui| {
                            let mirror = 0x00078000 + (self.index * 16);
                            let mut mirror_str = format!("{mirror:08x}");
                            ui.add_enabled(
                                false,
                                TextEdit::singleline(&mut mirror_str).horizontal_align(Align::Max),
                            );
                        });
                    });
                });
            let resource = VramResource::character(self.sim_id, self.palette, self.index);
            let image = Image::new(resource.to_uri())
                .maintain_aspect_ratio(true)
                .tint(Color32::RED)
                .texture_options(TextureOptions::NEAREST);
            ui.add(image);
            ui.section("Colors", |ui| {
                ui.horizontal(|ui| {
                    ui.label("Palette");
                    ComboBox::from_id_salt("palette")
                        .selected_text(self.palette.to_string())
                        .width(ui.available_width())
                        .show_ui(ui, |ui| {
                            for palette in VramPalette::values() {
                                ui.selectable_value(
                                    &mut self.palette,
                                    palette,
                                    palette.to_string(),
                                );
                            }
                        });
                });
                TableBuilder::new(ui)
                    .columns(Column::remainder(), 4)
                    .body(|mut body| {
                        body.row(30.0, |mut row| {
                            for index in 0..4 {
                                let resource =
                                    VramResource::palette_color(self.sim_id, self.palette, index);
                                row.col(|ui| {
                                    let rect = ui.available_rect_before_wrap();
                                    let scale = rect.height() / rect.width();
                                    let rect = rect.scale_from_center2(Vec2::new(scale, 1.0));
                                    let image = Image::new(resource.to_uri())
                                        .tint(Color32::RED)
                                        .fit_to_exact_size(rect.max - rect.min);
                                    ui.put(rect, image);
                                });
                            }
                        });
                    });
            });
            ui.section("Display", |ui| {
                ui.horizontal(|ui| {
                    ui.label("Scale");
                    ui.spacing_mut().slider_width = ui.available_width();
                    let slider = Slider::new(&mut self.scale, 1.0..=10.0)
                        .step_by(1.0)
                        .show_value(false);
                    ui.add(slider);
                });
                ui.checkbox(&mut self.show_grid, "Show grid");
            });
        });
    }

    fn show_chardata(&mut self, ui: &mut Ui) {
        let start_pos = ui.cursor().min;
        let resource = VramResource::character_data(self.sim_id, self.palette);
        let image = Image::new(resource.to_uri())
            .fit_to_original_size(self.scale)
            .tint(Color32::RED)
            .texture_options(TextureOptions::NEAREST)
            .sense(Sense::click());
        let res = ui.add(image);
        if res.clicked() {
            if let Some(click_pos) = res.interact_pointer_pos() {
                let fixed_pos = (click_pos - start_pos) / self.scale;
                let x = (fixed_pos.x / 8.0) as usize;
                let y = (fixed_pos.y / 8.0) as usize;
                self.index = (y * 16) + x;
                self.index_str = self.index.to_string();
            }
        }
        let painter = ui.painter_at(res.rect);
        if self.show_grid {
            let stroke = ui.style().visuals.widgets.noninteractive.fg_stroke;
            for x in (1..16).map(|i| (i as f32) * 8.0 * self.scale) {
                let p1 = res.rect.min + (x, 0.0).into();
                let p2 = res.rect.min + (x, 128.0 * 8.0 * self.scale).into();
                painter.line(vec![p1, p2], stroke);
            }
            for y in (1..128).map(|i| (i as f32) * 8.0 * self.scale) {
                let p1 = res.rect.min + (0.0, y).into();
                let p2 = res.rect.min + (16.0 * 8.0 * self.scale, y).into();
                painter.line(vec![p1, p2], stroke);
            }
        }
        // draw box around selected
        let x1 = (self.index % 16) as f32 * 8.0 * self.scale;
        let x2 = x1 + (8.0 * self.scale);
        let y1 = (self.index / 16) as f32 * 8.0 * self.scale;
        let y2 = y1 + (8.0 * self.scale);
        painter.line(
            vec![
                (res.rect.min + (x1, y1).into()),
                (res.rect.min + (x2, y1).into()),
                (res.rect.min + (x2, y2).into()),
                (res.rect.min + (x1, y2).into()),
                (res.rect.min + (x1, y1).into()),
            ],
            ui.style().visuals.widgets.active.fg_stroke,
        );
    }
}

impl AppWindow for CharacterDataWindow {
    fn viewport_id(&self) -> ViewportId {
        ViewportId::from_hash_of(format!("chardata-{}", self.sim_id))
    }

    fn sim_id(&self) -> SimId {
        self.sim_id
    }

    fn initial_viewport(&self) -> ViewportBuilder {
        ViewportBuilder::default()
            .with_title(format!("Character Data ({})", self.sim_id))
            .with_inner_size((640.0, 480.0))
    }

    fn show(&mut self, ctx: &Context) {
        CentralPanel::default().show(ctx, |ui| {
            ui.horizontal_top(|ui| {
                StripBuilder::new(ui)
                    .size(Size::relative(0.3))
                    .size(Size::remainder())
                    .horizontal(|mut strip| {
                        strip.cell(|ui| {
                            ScrollArea::vertical().show(ui, |ui| self.show_form(ui));
                        });
                        strip.cell(|ui| {
                            ScrollArea::both().show(ui, |ui| self.show_chardata(ui));
                        });
                    });
            });
        });
    }
}

trait UiExt {
    fn section(&mut self, title: impl Into<String>, add_contents: impl FnOnce(&mut Ui));
}

impl UiExt for Ui {
    fn section(&mut self, title: impl Into<String>, add_contents: impl FnOnce(&mut Ui)) {
        let mut frame = Frame::group(self.style());
        frame.outer_margin.top += 10.0;
        frame.inner_margin.top += 2.0;
        let res = frame.show(self, add_contents);
        let text = RichText::new(title).background_color(self.style().visuals.panel_fill);
        let old_rect = res.response.rect;
        let mut text_rect = old_rect;
        text_rect.min.x += 6.0;
        let new_rect = self
            .allocate_new_ui(UiBuilder::new().max_rect(text_rect), |ui| ui.label(text))
            .response
            .rect;
        self.allocate_space((old_rect.max - new_rect.max) - (old_rect.min - new_rect.min));
    }
}
